"""
spbla C API Python bridge.

Wraps native C API details for accessing
this functionality via Python CTypes library.

Functionality:
- Flags wrapping
- Functions definitions
- Error checking
"""

import ctypes

__all__ = [
    "load_and_configure",
    "get_init_hints",
    "get_build_hints",
    "get_sub_matrix_hints",
    "get_transpose_hints",
    "get_reduce_hints",
    "get_kronecker_hints",
    "get_mxm_hints",
    "get_ewiseadd_hints",
    "check"
]

_hint_no = 0
_hint_cpu_backend = 1
_hint_cuda_backend = 2
_hint_opencl_backend = 4
_hint_gpu_mem_managed = 8
_hint_values_sorted = 16
_hint_accumulate = 32
_hint_relaxed_release = 64
_hint_log_error = 128
_hint_log_warning = 256
_hint_log_all = 512
_hint_no_duplicates = 1024
_hint_time_check = 2048

_backend_name_cpu = "cpu"
_backend_name_cuda = "cuda"
_backend_name_opencl = "opencl"


def get_log_hints(default=True, error=False, warning=False):
    hints = _hint_no

    if default:
        hints |= _hint_log_all
    if error:
        hints |= _hint_log_error
    if warning:
        hints |= _hint_log_warning

    return hints


def get_init_hints(backend_type: str):
    assert backend_type

    hints = _hint_relaxed_release

    if backend_type == _backend_name_cpu:
        hints |= _hint_cpu_backend
    elif backend_type == _backend_name_cuda:
        hints |= _hint_cuda_backend
    elif backend_type == _backend_name_opencl:
        hints |= _hint_opencl_backend

    return hints


def get_sub_matrix_hints(time_check):
    hints = _hint_no

    if time_check:
        hints |= _hint_time_check

    return hints


def get_transpose_hints(time_check):
    hints = _hint_no

    if time_check:
        hints |= _hint_time_check

    return hints


def get_reduce_hints(time_check):
    hints = _hint_no

    if time_check:
        hints |= _hint_time_check

    return hints


def get_kronecker_hints(time_check):
    hints = _hint_no

    if time_check:
        hints |= _hint_time_check

    return hints


def get_mxm_hints(is_accumulated, time_check):
    hints = _hint_no

    if is_accumulated:
        hints |= _hint_accumulate
    if time_check:
        hints |= _hint_time_check

    return hints


def get_ewiseadd_hints(time_check):
    hints = _hint_no

    if time_check:
        hints |= _hint_time_check

    return hints


def get_build_hints(is_sorted, no_duplicates):
    hints = _hint_no

    if is_sorted:
        hints |= _hint_values_sorted
    if no_duplicates:
        hints |= _hint_no_duplicates

    return hints


def load_and_configure(cubool_lib_path: str):
    lib = ctypes.cdll.LoadLibrary(cubool_lib_path)

    status_t = ctypes.c_uint
    index_t = ctypes.c_uint
    hints_t = ctypes.c_uint
    matrix_p = ctypes.c_void_p
    p_to_matrix_p = ctypes.POINTER(matrix_p)

    lib.spbla_SetupLogging.restype = status_t
    lib.spbla_SetupLogging.argtypes = [
        ctypes.POINTER(ctypes.c_char),
        hints_t
    ]

    lib.spbla_Initialize.restype = status_t
    lib.spbla_Initialize.argtypes = [
        hints_t
    ]

    lib.spbla_Finalize.restype = status_t
    lib.spbla_Finalize.argtypes = []

    lib.spbla_Matrix_New.restype = status_t
    lib.spbla_Matrix_New.argtypes = [
        p_to_matrix_p,
        ctypes.c_uint,
        ctypes.c_uint
    ]

    lib.spbla_Matrix_Free.restype = status_t
    lib.spbla_Matrix_Free.argtypes = [
        matrix_p
    ]

    lib.spbla_Matrix_Build.restype = status_t
    lib.spbla_Matrix_Build.argtypes = [
        matrix_p,
        ctypes.POINTER(ctypes.c_uint),
        ctypes.POINTER(ctypes.c_uint),
        ctypes.c_uint,
        hints_t
    ]

    lib.spbla_Matrix_SetElement.restype = status_t
    lib.spbla_Matrix_SetElement.argtypes = [
        matrix_p,
        ctypes.c_uint,
        ctypes.c_uint
    ]

    lib.spbla_Matrix_SetMarker.restype = status_t
    lib.spbla_Matrix_SetMarker.argtypes = [
        matrix_p,
        ctypes.POINTER(ctypes.c_char)
    ]

    lib.spbla_Matrix_Marker.restype = status_t
    lib.spbla_Matrix_Marker.argtypes = [
        matrix_p,
        ctypes.POINTER(ctypes.c_char),
        ctypes.POINTER(ctypes.c_uint)
    ]

    lib.spbla_Matrix_ExtractPairs.restype = status_t
    lib.spbla_Matrix_ExtractPairs.argtypes = [
        matrix_p,
        ctypes.POINTER(ctypes.c_uint),
        ctypes.POINTER(ctypes.c_uint),
        ctypes.POINTER(ctypes.c_uint)
    ]

    lib.spbla_Matrix_ExtractSubMatrix.restype = status_t
    lib.spbla_Matrix_ExtractSubMatrix.argtypes = [
        matrix_p,
        matrix_p,
        index_t,
        index_t,
        index_t,
        index_t,
        hints_t
    ]

    lib.spbla_Matrix_Duplicate.restype = status_t
    lib.spbla_Matrix_Duplicate.argtypes = [
        matrix_p,
        p_to_matrix_p
    ]

    lib.spbla_Matrix_Transpose.restype = status_t
    lib.spbla_Matrix_Transpose.argtypes = [
        matrix_p,
        matrix_p,
        hints_t
    ]

    lib.spbla_Matrix_Nrows.restype = status_t
    lib.spbla_Matrix_Nrows.argtype = [
        matrix_p,
        ctypes.POINTER(ctypes.c_uint)
    ]

    lib.spbla_Matrix_Ncols.restype = status_t
    lib.spbla_Matrix_Ncols.argtype = [
        matrix_p,
        ctypes.POINTER(ctypes.c_uint)
    ]

    lib.spbla_Matrix_Nvals.restype = status_t
    lib.spbla_Matrix_Nvals.spbla_Matrix_Reduce = [
        matrix_p,
        ctypes.POINTER(ctypes.c_size_t)
    ]

    lib.spbla_Matrix_Reduce.restype = status_t
    lib.spbla_Matrix_Reduce.argtype = [
        matrix_p,
        matrix_p,
        hints_t
    ]

    lib.spbla_Matrix_EWiseAdd.restype = status_t
    lib.spbla_Matrix_EWiseAdd.argtypes = [
        matrix_p,
        matrix_p,
        matrix_p,
        hints_t
    ]

    lib.spbla_MxM.restype = status_t
    lib.spbla_MxM.argtypes = [
        matrix_p,
        matrix_p,
        matrix_p,
        hints_t
    ]

    lib.spbla_Kronecker.restype = status_t
    lib.spbla_Kronecker.argtypes = [
        matrix_p,
        matrix_p,
        matrix_p,
        hints_t
    ]

    return lib


"""
/** Possible status codes that can be returned from spbla api */

/** Successful execution of the function */
SPBLA_STATUS_SUCCESS,

/** Generic error code */
SPBLA_STATUS_ERROR,

/** No cuda compatible device in the system */
SPBLA_STATUS_DEVICE_NOT_PRESENT,

/** Device side error */
SPBLA_STATUS_DEVICE_ERROR,

/** Failed to allocate memory on cpy or gpu side */
SPBLA_STATUS_MEM_OP_FAILED,

/** Passed invalid argument to some function */
SPBLA_STATUS_INVALID_ARGUMENT,

/** Call of the function is not possible for some context */
SPBLA_STATUS_INVALID_STATE

/** Failed to select supported backend for computations */
SPBLA_STATUS_BACKEND_ERROR

/** Some library feature is not implemented */
SPBLA_STATUS_NOT_IMPLEMENTED
"""
_status_codes_mappings = {
    0: "SPBLA_STATUS_SUCCESS",
    1: "SPBLA_STATUS_ERROR",
    2: "SPBLA_STATUS_DEVICE_NOT_PRESENT",
    3: "SPBLA_STATUS_DEVICE_ERROR",
    4: "SPBLA_STATUS_MEM_OP_FAILED",
    5: "SPBLA_STATUS_INVALID_ARGUMENT",
    6: "SPBLA_STATUS_INVALID_STATE",
    7: "SPBLA_STATUS_BACKEND_ERROR",
    8: "SPBLA_STATUS_NOT_IMPLEMENTED"
}

_success = 0


def check(status_code):
    if status_code != _success:
        raise Exception(_status_codes_mappings[status_code])



"""
Utils for accessing spbla features.

Features:
- Allows to setup logging to custom file with filter settings
- Allows to setup default log
- Allows to create default log file name (for user purposes)
"""

from . import wrapper
from . import bridge
import ctypes
import pathlib
import datetime

__all__ = [
    "setup_logger",
    "setup_default_logger",
    "get_default_log_name"
]


def get_default_log_name():
    """
    Creates default log name, composed from current datetime info.

    :return: String log file
    """
    return "spbla-" + datetime.datetime.now().strftime("%d-%m-%y--%H-%M-%S") + ".textlog"


def setup_logger(file_path: str, default=True, error=False, warning=False):
    """
    Allows to setup logging into user defined logging file.

    :param file_path: Full/relative path to the file to save logged messages
    :param default: Set in true to use default (all) log filter
    :param error: Set in true to log errors
    :param warning: Set in true to log warnings
    :return: None
    """

    status = wrapper.loaded_dll.spbla_SetupLogging(
        file_path.encode("utf-8"),
        ctypes.c_uint(bridge.get_log_hints(default, error, warning))
    )

    bridge.check(status)


def setup_default_logger():
    """
    Setups default logger, with automatically selected log file
    and default logged messages filter settings.

    :return:
    """

    here = pathlib.Path(__file__).parent
    log_path = here / get_default_log_name()

    setup_logger(str(log_path), default=True)



"""
spbla library state wrapper.
Provides global access points and initialization logic for library API.
"""

import os
import ctypes
import pathlib
from . import bridge

__all__ = [
    "singleton",
    "loaded_dll",
    "init_wrapper"
]

# Wrapper
singleton = None

# Exposed imported dll (owned by _lib_wrapper)
loaded_dll = None


def init_wrapper():
    global singleton
    global loaded_dll

    singleton = Wrapper()
    loaded_dll = singleton.loaded_dll


class Wrapper:
    """
    Instance of this class represents the main access point
    to the C spbla library. It stores:

     - Ctypes imported dll instance
     - Ctypes functions declarations
     - Import path

    At exit the library is properly finalized.
    Instance of the spbla library in released and the shared library is closed.
    """

    def __init__(self):
        self.loaded_dll = None
        self.backend = "default"

        try:
            # Try from config if present
            self.load_path = os.environ["SPBLA_PATH"]
        except KeyError:
            # Fallback to package directory
            source_path = pathlib.Path(__file__).resolve()
            self.load_path = Wrapper.__get_lib_path(source_path.parent)

        try:
            # Check, if user has backend-type preferences
            self.backend = str(os.environ["SPBLA_BACKEND"]).lower()
        except KeyError:
            pass

        assert self.load_path
        assert self.backend

        self.loaded_dll = bridge.load_and_configure(self.load_path)
        self.__setup_library()

    def __del__(self):
        self.__release_library()

    def __setup_library(self):
        status = self.loaded_dll.spbla_Initialize(ctypes.c_uint(bridge.get_init_hints(self.backend)))
        bridge.check(status)

    def __release_library(self):
        status = self.loaded_dll.spbla_Finalize()
        bridge.check(status)

    @classmethod
    def __get_lib_path(cls, prefix):
        for entry in os.listdir(prefix):
            if "spbla" in str(entry):
                return prefix / entry

        return None



"""
Entry point for `pyspbla` package.
pyspbla is python wrapper for spbla library.

The backend spbla compiled library is initialized automatically
on package import into runtime. State of the library is managed
by the wrapper class. All resources are unloaded automatically on
python runtime exit.

Exported primitives:
- matrix (sparse matrix of boolean values)

For more information refer to:
- spbla project: https://github.com/JetBrains-Research/spbla
- bug tracker: https://github.com/JetBrains-Research/spbla/issues

License:
- MIT
"""

from .wrapper import *
from .utils import *
from .matrix import *
from .io import *
from .gviz import *

# Setup global module state
init_wrapper()



from io import StringIO
from . import Matrix


__all__ = [
    "matrix_data_to_gviz",
    "matrices_data_to_gviz",
    "matrices_to_gviz"
]


DEFAULT_VERTEX_COLOR = "black"
DEFAULT_EDGE_COLOR = "black"
DEFAULT_LABEL = ""
DEFAULT_BASE_VERTEX = 0


def matrices_to_gviz(matrices: dict, **kwargs) -> str:
    """
    Export the labeled square matrices dictionary to the graph viz graph description script.
    All matrices must have the save shape.

    >>> name = "Test"                           # Displayed graph name
    >>> shape = (4, 4)                          # Adjacency matrices shape
    >>> colors = {"a": "red", "b": "green"}     # Colors per label
    >>>
    >>> a = Matrix.empty(shape=shape)        # Edges labeled as 'a'
    >>> a[0, 1] = True
    >>> a[1, 2] = True
    >>> a[2, 0] = True
    >>>
    >>> b = Matrix.empty(shape=shape)        # Edges labeled as 'b'
    >>> b[2, 3] = True
    >>> b[3, 2] = True
    >>>
    >>> print(matrices_to_gviz(matrices={"a": a, "b": b}, graph_name=name, edge_colors=colors))
    '
        digraph G {
            graph [label=Test];
            node [color=black];
            0 -> 1 [label=a,color=red];
            1 -> 2 [label=a,color=red];
            2 -> 0 [label=a,color=red];
            2 -> 3 [label=b,color=green];
            3 -> 2 [label=b,color=green];
        }
    '

    :param matrices: The dictionary with matrix label mapping to the pyspbla Matrix instance
    :param kwargs: Optional arguments to tweak gviz graph appearance
        :param graph_name: Name of the graph
        :param base_vertex: Index of the base vertex (by default is 0)
        :param vertex_color: Color of the vertices (use \" - \" to pass html style hex colors)
        :param edges_colors: Dictionary with mapping matrix label to its edges colors (use \" - \" to pass html style hex colors)

    :return: Text script gviz graph representation
    """

    assert len(matrices) > 0

    base = next(iter(matrices.values()))
    shape = base.shape

    for m in matrices.values():
        assert m.shape == shape

    matrices_data = dict()

    for key, value in matrices.items():
        assert isinstance(value, Matrix)
        matrices_data[key] = value.to_lists()

    return matrices_data_to_gviz(shape=shape, matrices=matrices_data, **kwargs)


def matrix_data_to_gviz(shape, rows, cols, **kwargs) -> str:
    check_len_rows_cols(rows, cols)
    check_shape(shape)

    _result = StringIO()
    args = unboxing_kwargs([], **kwargs)
    _result.write("digraph G {\n")
    _result.write(args["graph_name"])
    _result.write(args["vertex_color"])
    n = shape[0]

    used_vertex = [False] * n
    _result.write(build_edges(rows, cols, args["base_vertex"], args["label"], args["edge_color"], used_vertex))

    for i, v in enumerate(used_vertex):
        if not v:
            _result.write(f'{i + args["base_vertex"]};\n')

    _result.write("}\n")
    result = _result.getvalue()
    _result.close()

    return result


def matrices_data_to_gviz(shape, matrices: dict, **kwargs):
    check_shape(shape)
    for name in matrices.keys():
        check_len_rows_cols(matrices[name][0], matrices[name][1])

    _result = StringIO()
    args = unboxing_kwargs(matrices.keys(), **kwargs)
    _result.write("digraph G {\n")
    _result.write(args["graph_name"])
    _result.write(args["vertex_color"])
    n = shape[0]

    used_vertex = [False] * n
    for name in matrices.keys():
        rows = matrices[name][0]
        cols = matrices[name][1]
        label = name
        edge_color = args["edge_colors"].get(name)
        _result.write(build_edges(rows, cols, args["base_vertex"], label, edge_color, used_vertex))

    for i, v in enumerate(used_vertex):
        if not v:
            _result.write(f'{i + args["base_vertex"]};\n')

    _result.write("}\n")
    result = _result.getvalue()
    _result.close()

    return result


def build_edges(rows, cols, base_vertex, label, edge_color, used_vertex):
    _result = StringIO()

    for i, j in zip(rows, cols):
        _result.write(f'{i + base_vertex} -> {j + base_vertex} [label={label},color={edge_color}];\n')
        used_vertex[i] = True
        used_vertex[j] = True

    result = _result.getvalue()
    _result.close()

    return result


def unboxing_kwargs(labels_set, **kwargs) -> dict:
    _graph_name = kwargs.get("graph_name")
    _vertex_color = kwargs.get("vertex_color")
    _edge_color = kwargs.get("edge_color")
    _base_vertex = kwargs.get("base_vertex")
    _label = kwargs.get("label")
    edge_colors = kwargs.get("edge_colors")

    graph_name = "" if _graph_name is None else f'graph [label={_graph_name}];\n'
    vertex_color = f'node [color={DEFAULT_VERTEX_COLOR if _vertex_color is None else _vertex_color}];\n'
    edge_color = DEFAULT_EDGE_COLOR if _edge_color is None else _edge_color
    base_vertex = DEFAULT_BASE_VERTEX if _base_vertex is None else _base_vertex
    label = DEFAULT_LABEL if _label is None else _label

    presenter_edge_colors = edge_colors.keys()
    for l in labels_set:
        if l not in presenter_edge_colors:
            edge_colors[l] = DEFAULT_EDGE_COLOR

    result = dict()
    result["graph_name"] = graph_name
    result["vertex_color"] = vertex_color
    result["edge_color"] = edge_color
    result["label"] = label
    result["base_vertex"] = base_vertex
    result["edge_colors"] = edge_colors

    return result


def check_shape(shape: list):
    if shape[0] != shape[1]:
        raise Exception("Matrix must be square")


def check_len_rows_cols(rows, cols):
    if len(rows) != len(cols):
        raise Exception("Rows and cols arrays must have equal size")



"""
Matrix primitive.
"""

import ctypes
import random

from . import wrapper
from . import bridge


__all__ = [
    "Matrix"
]


class Matrix:
    """
    Wrapper for spbla Sparse boolean matrix type.

    Matrix class supports all spbla C API Matrix functions.
    Also Matrix class provides additional fancy functions/operators for better user experience.

    Matrix creation:
    - empty
    - from lists data
    - random generated

    Matrix operations:
    - mxm
    - ewiseadd
    - kronecker
    - reduce
    - transpose
    - matrix extraction

    Matrix functions:
    - to string
    - values iterating
    - equality check

    Debug features:
    - String markers
    """

    __slots__ = ["hnd"]

    def __init__(self, hnd):
        self.hnd = hnd

    def __del__(self):
        bridge.check(wrapper.loaded_dll.spbla_Matrix_Free(self.hnd))

    @classmethod
    def empty(cls, shape):
        """
        Creates empty matrix of specified `shape`.

        :param shape: Pair with two values with rows and cols count of the matrix
        :return: Created empty matrix
        """

        hnd = ctypes.c_void_p(0)

        nrows = shape[0]
        ncols = shape[1]

        status = wrapper.loaded_dll.spbla_Matrix_New(
            ctypes.byref(hnd), ctypes.c_uint(nrows), ctypes.c_uint(ncols)
        )

        bridge.check(status)

        return Matrix(hnd)

    @classmethod
    def from_lists(cls, shape, rows, cols, is_sorted=False, no_duplicates=False):
        """
        Build matrix from provided `shape` and non-zero values data.

        >>> matrix = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True, no_duplicates=True)
        >>> print(matrix)
        '
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   1   .   .   . |   3
                0   1   2   3
        '

        :param shape: Matrix shape
        :param rows: List with row indices
        :param cols: List with column indices
        :param is_sorted: True if values are sorted in row-col order
        :param no_duplicates: True if provided values has no duplicates
        :return: Created matrix filled with data
        """

        out = cls.empty(shape)
        out.build(rows, cols, is_sorted=is_sorted, no_duplicates=no_duplicates)
        return out

    @classmethod
    def generate(cls, shape, density: float):
        """
        Generate matrix of the specified shape with desired values density.

        >>> matrix = Matrix.generate(shape=(4, 4), density=0.5)
        >>> print(matrix)
        '
                0   1   2   3
          0 |   .   1   .   . |   0
          1 |   .   .   1   1 |   1
          2 |   1   1   .   . |   2
          3 |   .   1   .   . |   3
                0   1   2   3
        '

        :param shape: Matrix shape to generate
        :param density: Matrix values density, must be within [0, 1] bounds
        :return: Generated matrix
        """

        density = min(1.0, max(density, 0.0))
        nvals_max = shape[0] * shape[1]
        nvals_to_gen = int(nvals_max * density)

        m, n = shape
        rows, cols = list(), list()

        for i in range(nvals_to_gen):
            rows.append(random.randrange(0, m))
            cols.append(random.randrange(0, n))

        return Matrix.from_lists(shape=shape, rows=rows, cols=cols, is_sorted=False, no_duplicates=False)

    def build(self, rows, cols, is_sorted=False, no_duplicates=False):
        """
        Build sparse matrix of boolean values from provided arrays of non-zero rows and columns.

        >>> matrix = Matrix.empty(shape=(4,4))
        >>> matrix.build([0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True, no_duplicates=True)
        >>> print(matrix)
        '
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   1   .   .   . |   3
                0   1   2   3
        '

        :param rows: Array of values rows indices
        :param cols: Array of values column indices
        :param is_sorted: True if values are sorted in row-col order
        :param no_duplicates: True if provided values has no duplicates
        :return:
        """

        if len(rows) != len(cols):
            raise Exception("Rows and cols arrays must have equal size")

        nvals = len(rows)
        t_rows = (ctypes.c_uint * len(rows))(*rows)
        t_cols = (ctypes.c_uint * len(cols))(*cols)

        status = wrapper.loaded_dll.spbla_Matrix_Build(
            self.hnd, t_rows, t_cols,
            ctypes.c_uint(nvals),
            ctypes.c_uint(bridge.get_build_hints(is_sorted, no_duplicates))
        )

        bridge.check(status)

    def dup(self):
        """
        Creates new matrix instance, the exact copy of the `self`

        >>> a = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True, no_duplicates=True)
        >>> b = a.dup()
        >>> b[3, 3] = True
        >>> print(a, b, sep="")
        '
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   1   .   .   . |   3
                0   1   2   3
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   1   .   .   1 |   3
                0   1   2   3
        '

        :return: New matrix instance with `self` copied data
        """

        hnd = ctypes.c_void_p(0)

        status = wrapper.loaded_dll.spbla_Matrix_Duplicate(
            self.hnd, ctypes.byref(hnd)
        )

        bridge.check(status)
        return Matrix(hnd)

    def transpose(self, time_check=False):
        """
        Creates new transposed `self` matrix.

        >>> a = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True, no_duplicates=True)
        >>> b = a.transpose()
        >>> print(a, b, sep="")
        '
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   1   .   .   . |   3
                0   1   2   3
                0   1   2   3
          0 |   1   .   .   1 |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   .   .   .   . |   3
                0   1   2   3
        '

        :param time_check: Pass True to measure and log elapsed time of the operation
        :return: New matrix instance with `self` transposed data
        """

        shape = (self.ncols, self.nrows)
        out = Matrix.empty(shape)

        status = wrapper.loaded_dll.spbla_Matrix_Transpose(
            out.hnd,
            self.hnd,
            ctypes.c_uint(bridge.get_transpose_hints(time_check=time_check))
        )

        bridge.check(status)
        return out

    def set_marker(self, marker: str):
        """
        Sets to the matrix specific debug string marker.
        This marker will appear in the log messages as string identifier of the matrix.

        >>> a = Matrix.empty(shape=(4, 4))
        >>> print(a.marker)
        '0x1a767b0'
        >>> a.set_marker("meow")
        >>> print(a.marker)
        'meow (0x1a767b0)'

        :param marker: String marker to set
        :return: None
        """

        assert marker is not None

        status = wrapper.loaded_dll.spbla_Matrix_SetMarker(
            self.hnd, marker.encode("utf-8")
        )

        bridge.check(status)
        return None

    @property
    def marker(self):
        """
        Allows to get matrix debug string marker.

        >>> a = Matrix.empty(shape=(4, 4))
        >>> print(a.marker)
        '0x1a767b0'
        >>> a.set_marker("meow")
        >>> print(a.marker)
        'meow (0x1a767b0)'

        :return: String matrix marker.
        """

        size = ctypes.c_uint(0)
        status = wrapper.loaded_dll.spbla_Matrix_Marker(
            self.hnd, ctypes.POINTER(ctypes.c_char)(), ctypes.byref(size)
        )

        bridge.check(status)

        c_buffer = (ctypes.c_char * int(size.value))()
        status = wrapper.loaded_dll.spbla_Matrix_Marker(
            self.hnd, c_buffer, ctypes.byref(size)
        )

        bridge.check(status)
        return c_buffer.value.decode("utf-8")

    @property
    def nrows(self) -> int:
        """
        Query number of rows of the `self` matrix.
        :return: Number of rows
        """

        result = ctypes.c_uint(0)

        status = wrapper.loaded_dll.spbla_Matrix_Nrows(
            self.hnd, ctypes.byref(result)
        )

        bridge.check(status)
        return int(result.value)

    @property
    def ncols(self) -> int:
        """
        Query number of columns of the `self` matrix.
        :return: Number of columns
        """

        result = ctypes.c_uint(0)

        status = wrapper.loaded_dll.spbla_Matrix_Ncols(
            self.hnd, ctypes.byref(result)
        )

        bridge.check(status)
        return int(result.value)

    @property
    def nvals(self) -> int:
        """
        Query number of non-zero values of the `self` matrix.
        :return: Number of non-zero values
        """

        result = ctypes.c_uint(0)

        status = wrapper.loaded_dll.spbla_Matrix_Nvals(
            self.hnd, ctypes.byref(result)
        )

        bridge.check(status)
        return int(result.value)

    @property
    def shape(self) -> (int, int):
        """
        Query shape of `self` matrix as (nrows, ncols) tuple.
        :return: Return tuple of (nrows, ncols)
        """

        return self.nrows, self.ncols

    def to_lists(self):
        """
        Read matrix data as lists of `rows` and `clos` indices.

        >>> a = Matrix.empty(shape=(4, 4))
        >>> a[0, 0] = True
        >>> a[1, 3] = True
        >>> a[1, 0] = True
        >>> a[2, 2] = True
        >>> rows, cols = a.to_lists()
        >>> print(list(rows), list(cols))
        '[0, 1, 1, 2] [0, 0, 3, 2]'

        :return: Pair with `rows` and `cols` lists
        """

        count = self.nvals

        rows = (ctypes.c_uint * count)()
        cols = (ctypes.c_uint * count)()
        nvals = ctypes.c_uint(count)

        status = wrapper.loaded_dll.spbla_Matrix_ExtractPairs(
            self.hnd, rows, cols, ctypes.byref(nvals)
        )

        bridge.check(status)

        return rows, cols

    def to_list(self):
        """
        Read matrix values as list of (i,j) pairs.

        >>> a = Matrix.empty(shape=(4, 4))
        >>> a[0, 0] = True
        >>> a[1, 3] = True
        >>> a[1, 0] = True
        >>> a[2, 2] = True
        >>> vals = a.to_list()
        >>> print(vals)
        '[(0, 0), (1, 0), (1, 3), (2, 2)]'

        :return: List of (i, j) pairs
        """

        I, J = self.to_lists()
        return list(zip(I, J))

    def to_string(self, width=3):
        """
        Return a string representation of the matrix.

        >>> matrix = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True, no_duplicates=True)
        >>> print(matrix)
        '
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   1   . |   2
          3 |   1   .   .   . |   3
                0   1   2   3
        '

        :param width: Width of the field in chars where to put numbers of rows and columns
        :return: Matrix string representation
        """

        nrows = self.nrows
        ncols = self.ncols
        nvals = self.nvals
        rows, cols = self.to_lists()

        cell_empty = "."
        cell_filled = "1"
        cell_sep = " "
        format_str = "{:>%s}" % width

        header = format_str.format("") + "  " + cell_sep + "".join(format_str.format(j) + cell_sep for j in range(ncols))
        result = header + "\n"

        v = 0
        for i in range(nrows):
            line = format_str.format(i) + " |" + cell_sep
            for j in range(ncols):
                if v < nvals and rows[v] == i and cols[v] == j:
                    line += format_str.format(cell_filled) + cell_sep
                    v += 1
                else:
                    line += format_str.format(cell_empty) + cell_sep
            line += "| " + format_str.format(i) + "\n"
            result += line

        result += header + "\n"
        return result

    def extract_matrix(self, i, j, shape, out=None, time_check=False):
        """
        Extract a sub-matrix.

        >>> matrix = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True)
        >>> print(matrix.extract_matrix(0, 1, shape=(3, 3)))
        '
                0   1   2
          0 |   .   .   . |   0
          1 |   1   .   . |   1
          2 |   .   1   . |   2
                0   1   2
        '

        :param i: First row index to extract
        :param j: First column index to extract
        :param shape: Shape of the sub-matrix
        :param out: Optional matrix where to store result
        :param time_check: Pass True to measure and log elapsed time of the operation
        :return: Sub-matrix
        """

        if out is None:
            out = Matrix.empty(shape)

        status = wrapper.loaded_dll.spbla_Matrix_ExtractSubMatrix(
            out.hnd, self.hnd,
            ctypes.c_uint(i),
            ctypes.c_uint(j),
            ctypes.c_uint(shape[0]),
            ctypes.c_uint(shape[1]),
            ctypes.c_uint(bridge.get_sub_matrix_hints(time_check=time_check))
        )

        bridge.check(status)
        return out

    def mxm(self, other, out=None, accumulate=False, time_check=False):
        """
        Matrix-matrix multiplication in boolean semiring with "x = and" and "+ = or" operations.
        Returns `self` multiplied to `other` matrix.

        Pass optional `out` matrix to store result.
        Pass `accumulate`=True to sum the multiplication result with `out` matrix.

        >>> a = Matrix.from_lists((4, 4), [0, 1, 2], [2, 3, 0])
        >>> b = Matrix.from_lists((4, 4), [0, 1, 3], [2, 3, 0])
        >>> print(a.mxm(b, out=a, accumulate=True))
        '
                0   1   2   3
          0 |   .   .   1   . |   0
          1 |   1   .   .   1 |   1
          2 |   1   .   1   . |   2
          3 |   .   .   .   . |   3
                0   1   2   3
        '

        :param other: Input matrix for multiplication
        :param out: Optional out matrix to store result
        :param accumulate: Set in true to accumulate the result with `out` matrix
        :param time_check: Pass True to measure and log elapsed time of the operation
        :return: Matrix-matrix multiplication result (with possible accumulation to `out` if provided)
        """

        if out is None:
            shape = (self.nrows, other.ncols)
            out = Matrix.empty(shape)
            accumulate = False

        status = wrapper.loaded_dll.spbla_MxM(
            out.hnd,
            self.hnd,
            other.hnd,
            ctypes.c_uint(bridge.get_mxm_hints(is_accumulated=accumulate, time_check=time_check))
        )

        bridge.check(status)
        return out

    def kronecker(self, other, time_check=False):
        """
        Matrix-matrix kronecker product with boolean "x = and" operation.
        Returns kronecker product of `self` and `other` matrices.

        >>> a = Matrix.from_lists((3, 3), [0, 0, 1, 2, 2], [0, 2, 1, 0, 2])
        >>> b = Matrix.from_lists((3, 3), [0, 1, 1, 2], [1, 0, 2, 1])
        >>> print(a.kronecker(b))
        '
                0   1   2   3   4   5   6   7   8
          0 |   .   1   .   .   .   .   .   1   . |   0
          1 |   1   .   1   .   .   .   1   .   1 |   1
          2 |   .   1   .   .   .   .   .   1   . |   2
          3 |   .   .   .   .   1   .   .   .   . |   3
          4 |   .   .   .   1   .   1   .   .   . |   4
          5 |   .   .   .   .   1   .   .   .   . |   5
          6 |   .   1   .   .   .   .   .   1   . |   6
          7 |   1   .   1   .   .   .   1   .   1 |   7
          8 |   .   1   .   .   .   .   .   1   . |   8
                0   1   2   3   4   5   6   7   8
        '

        :param other: Input matrix
        :param time_check: Pass True to measure and log elapsed time of the operation
        :return: Matrices kronecker product matrix
        """

        shape = (self.nrows * other.nrows, self.ncols * other.ncols)
        out = Matrix.empty(shape)

        status = wrapper.loaded_dll.spbla_Kronecker(
            out.hnd,
            self.hnd,
            other.hnd,
            ctypes.c_uint(bridge.get_kronecker_hints(time_check=time_check))
        )

        bridge.check(status)
        return out

    def ewiseadd(self, other, time_check=False):
        """
        Element-wise matrix-matrix addition with boolean "+ = or" operation.
        Returns element-wise sum of `self` and `other` matrix.

        >>> a = Matrix.from_lists((4, 4), [0, 1, 2], [2, 3, 0])
        >>> b = Matrix.from_lists((4, 4), [0, 1, 3], [2, 3, 0])
        >>> print(a.ewiseadd(b))
        '
                0   1   2   3
          0 |   .   .   1   . |   0
          1 |   .   .   .   1 |   1
          2 |   1   .   .   . |   2
          3 |   1   .   .   . |   3
                0   1   2   3
        '

        :param other: Input matrix to sum
        :param time_check: Pass True to measure and log elapsed time of the operation
        :return: Element-wise matrix-matrix sum
        """

        shape = (self.nrows, self.ncols)
        out = Matrix.empty(shape)

        status = wrapper.loaded_dll.spbla_Matrix_EWiseAdd(
            out.hnd,
            self.hnd,
            other.hnd,
            ctypes.c_uint(bridge.get_ewiseadd_hints(time_check=time_check))
        )

        bridge.check(status)
        return out

    def reduce(self, time_check=False):
        """
        Reduce matrix to vector with boolean "+ = or" operation.
        Return `self` reduced matrix.

        >>> matrix = Matrix.from_lists((4, 4), [0, 1, 2, 2], [0, 1, 0, 2])
        >>> print(matrix.reduce())
        '
                  0
            0 |   1 |   0
            1 |   1 |   1
            2 |   1 |   2
            3 |   . |   3
                  0
        '

        :param time_check: Pass True to measure and log elapsed time of the operation
        :return: Reduced matrix (matrix with M x 1 shape)
        """

        shape = (self.nrows, 1)
        out = Matrix.empty(shape)

        status = wrapper.loaded_dll.spbla_Matrix_Reduce(
            out.hnd,
            self.hnd,
            ctypes.c_uint(bridge.get_reduce_hints(time_check=time_check))
        )

        bridge.check(status)
        return out

    def equals(self, other) -> bool:
        """
        Compare two matrices. Returns true if they are equal.
        todo: Add this method into C API

        :param other: Other matrix to compare
        :return: True if matrices are equal
        """

        if not self.shape == other.shape:
            return False
        if not self.nvals == other.nvals:
            return False

        self_rows, self_cols = self.to_lists()
        other_rows, other_cols = other.to_lists()

        for i in range(len(self_rows)):
            if self_rows[i] != other_rows[i]:
                return False
        for i in range(len(self_cols)):
            if self_cols[i] != other_cols[i]:
                return False

        return True

    def __str__(self):
        return self.to_string()

    def __iter__(self):
        """
        Iterate over (i, j) tuples of the matrix values.

        >>> matrix = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True)
        >>> print(list(iter(matrix)))
        '[(0, 0), (1, 1), (2, 2), (3, 0)]'

        :return: Matrix tuples iterator
        """

        rows, cols = self.to_lists()
        return zip(rows, cols)

    def __getitem__(self, item):
        """
        Extract sub-matrix from `self`.
        Supported only tuple `item` with two slices. Step in slices is not supported.

        >>> matrix = Matrix.from_lists((4, 4), [0, 1, 2, 3], [0, 1, 2, 0], is_sorted=True)
        >>> print(matrix[0:3, 1:])
        '
                0   1   2
          0 |   .   .   . |   0
          1 |   1   .   . |   1
          2 |   .   1   . |   2
                0   1   2
        '

        :param item: Tuple of two slices for rows and cols regions
        :return: Extracted sub-matrix
        """

        if isinstance(item, tuple):
            first = item[0]
            second = item[1]

            if isinstance(first, slice) and isinstance(second, slice):
                i = first.start
                iend = first.stop

                j = second.start
                jend = second.stop

                assert first.step is None
                assert second.step is None

                if i is None:
                    i = 0
                if j is None:
                    j = 0

                assert 0 <= i < self.nrows
                assert 0 <= j < self.ncols

                if iend is None:
                    iend = self.nrows
                if jend is None:
                    jend = self.ncols

                shape = (iend - i, jend - j)
                return self.extract_matrix(i, j, shape)

        raise Exception("Invalid matrix slicing")

    def __setitem__(self, key, value):
        """
        Sets Sets specified `key` = (i, j) value of the matrix to True.

        >>> matrix = Matrix.empty(shape=(4, 4))
        >>> matrix[0, 0] = True
        >>> matrix[1, 1] = True
        >>> matrix[2, 3] = True
        >>> matrix[3, 1] = True
        >>> print(matrix)
        '
                0   1   2   3
          0 |   1   .   .   . |   0
          1 |   .   1   .   . |   1
          2 |   .   .   .   1 |   2
          3 |   .   1   .   . |   3
                0   1   2   3
        '

        :param key: (i, j) pair to set matrix element in True
        :param value: Must be True always
        :return:
        """

        assert value is True

        if isinstance(key, tuple):
            i = key[0]
            j = key[1]

            status = wrapper.loaded_dll.spbla_Matrix_SetElement(
                self.hnd,
                ctypes.c_uint(i),
                ctypes.c_uint(j)
            )

            bridge.check(status)
            return

        raise Exception("Invalid item assignment")




"""
IO operations for exporting/importing mtx data.
Provides features to import/export data or pyspbla matrix in mtx format.
"""

from . import Matrix


__all__ = [
    "read_mtx_file",
    "write_mtx_file",
    "import_matrix_from_mtx",
    "export_matrix_to_mtx"
]


def read_mtx_file(path: str):
    """
    Reads mtx file data.

    :param path: Path and name of the file with data
    :return: shape of the matrix, rows data, columns data, number of values
    """

    with open(str(path), 'r') as file:
        line = file.readline()
        while line[0].startswith("#"):
            line = file.readline()

        m, n, nvals = map(int, line.split())
        rows = list()
        cols = list()
        for k in range(nvals):
            line = file.readline()
            if line.startswith("#"):
                continue

            i, j = map(int, line.split())
            rows.append(i)
            cols.append(j)

    return (m, n), rows, cols, nvals


def write_mtx_file(path: str, shape, rows, cols, nvals):
    """
    Save data to mtx file.

    :param path: Path and file name of the file to save data
    :param shape: Matrix shape as tuple
    :param rows: Rows data
    :param cols: Columns data
    :param nvals: Number of values
    :return: None
    """

    assert len(rows) == nvals
    assert len(cols) == nvals

    with open(str(path), "w") as file:
        file.write("# pyspbla sparse boolean matrix\n")
        file.write(f"{shape[0]} {shape[1]} {nvals}\n")
        for i in range(nvals):
            file.write(f"{rows[i]} {cols[i]}\n")


def import_matrix_from_mtx(path: str):
    """
    Read matrix from file in the mtx format.

    :param path: Path and name of the file with matrix
    :return: Matrix created from data
    """

    shape, rows, cols, nvals = read_mtx_file(path)
    return Matrix.from_lists(shape=shape, rows=rows, cols=cols, is_sorted=False, no_duplicates=False)


def export_matrix_to_mtx(path: str, matrix: Matrix):
    """
    Save matrix to mtx file.

    :param path: Path and file name of the file to save data
    :param matrix: Matrix to export
    :return: None
    """

    rows, cols = matrix.to_lists()
    nvals = matrix.nvals
    shape = matrix.shape
    write_mtx_file(path=path, shape=shape, rows=rows, cols=cols, nvals=nvals)


import os
import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
with open(HERE / "README.md", "r", encoding="utf-8") as fh:
    README = fh.read()

# Find lib name
SOURCE_PATH = pathlib.Path(__file__).resolve()
LIB_NAME = None

for entry in os.listdir(SOURCE_PATH.parent / "pyspbla"):
    if "spbla" in str(entry):
        LIB_NAME = entry
        break

setuptools.setup(
    name="pyspbla",
    version="1.0.0",
    author="Egor Orachyov",
    author_email="egororachyov@gmail.com",
    license="MIT",
    description="spbla library python bindings.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/JetBrains-Research/spbla",
    project_urls={
        "spbla project": "https://github.com/JetBrains-Research/spbla",
        "Bug Tracker": "https://github.com/JetBrains-Research/spbla/issues"
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Environment :: GPU",
        "Environment :: GPU :: NVIDIA CUDA",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ],
    keywords=[
        "python",
        "cplusplus",
        "sparse-matrix",
        "linear-algebra",
        "graph-analysis",
        "graph-algorithms",
        "graphblas",
        "nvidia-cuda",
        "opencl"
    ],
    packages=["pyspbla"],
    package_dir={'': '.'},
    package_data={'': [LIB_NAME]},
    python_requires=">=3.0",
    include_package_data=True
)



from .wrapper import *
from .utils import *
from .matrix import *
from .io import *

# Setup global module state
init_wrapper()

print("**pyspbla project**")
print("pyspbla is a python wrapper for a spbla library.")
print("See https://github.com/JetBrains-Research/spbla for more information.")
